package org.eclipse.osgi.util;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.io.InputStream;
import java.util.StringTokenizer;
import org.osgi.framework.BundleException;
import org.eclipse.osgi.framework.internal.core.Tokenizer;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

public class ManifestElement
{
    private static final String MANIFEST_INVALID_SPACE = "The manifest line \"{0}\" has an invalid leading space '' '' character.";
    private static final String MANIFEST_INVALID_LINE_NOCOLON = "The manifest line \"{0}\" is invalid; it has no colon '':'' character after the header key.";
    private static final String MANIFEST_INVALID_HEADER_EXCEPTION = "Invalid manifest header {0}: \"{1}\"";
    protected String value;
    protected String[] valueComponents;
    protected Hashtable attributes;
    protected Hashtable directives;
    
    protected ManifestElement() {
    }
    
    public String getValue() {
        return this.value;
    }
    
    public String[] getValueComponents() {
        return this.valueComponents;
    }
    
    public String getAttribute(final String key) {
        return this.getTableValue(this.attributes, key);
    }
    
    public String[] getAttributes(final String key) {
        return this.getTableValues(this.attributes, key);
    }
    
    public Enumeration getKeys() {
        return this.getTableKeys(this.attributes);
    }
    
    protected void addAttribute(final String key, final String value) {
        this.attributes = this.addTableValue(this.attributes, key, value);
    }
    
    public String getDirective(final String key) {
        return this.getTableValue(this.directives, key);
    }
    
    public String[] getDirectives(final String key) {
        return this.getTableValues(this.directives, key);
    }
    
    public Enumeration getDirectiveKeys() {
        return this.getTableKeys(this.directives);
    }
    
    protected void addDirective(final String key, final String value) {
        this.directives = this.addTableValue(this.directives, key, value);
    }
    
    private String getTableValue(final Hashtable table, final String key) {
        if (table == null) {
            return null;
        }
        final Object result = table.get(key);
        if (result == null) {
            return null;
        }
        if (result instanceof String) {
            return (String)result;
        }
        final ArrayList valueList = (ArrayList)result;
        return valueList.get(valueList.size() - 1);
    }
    
    private String[] getTableValues(final Hashtable table, final String key) {
        if (table == null) {
            return null;
        }
        final Object result = table.get(key);
        if (result == null) {
            return null;
        }
        if (result instanceof String) {
            return new String[] { (String)result };
        }
        final ArrayList valueList = (ArrayList)result;
        return valueList.toArray(new String[valueList.size()]);
    }
    
    private Enumeration getTableKeys(final Hashtable table) {
        if (table == null) {
            return null;
        }
        return table.keys();
    }
    
    private Hashtable addTableValue(Hashtable table, final String key, final String value) {
        if (table == null) {
            table = new Hashtable(7);
        }
        final Object curValue = table.get(key);
        if (curValue != null) {
            ArrayList newList;
            if (curValue instanceof ArrayList) {
                newList = (ArrayList)curValue;
            }
            else {
                newList = new ArrayList(5);
                newList.add(curValue);
            }
            newList.add(value);
            table.put(key, newList);
        }
        else {
            table.put(key, value);
        }
        return table;
    }
    
    public static ManifestElement[] parseHeader(final String header, final String value) throws BundleException {
        if (value == null) {
            return null;
        }
        final ArrayList headerElements = new ArrayList(10);
        final Tokenizer tokenizer = new Tokenizer(value);
        while (true) {
            String next = tokenizer.getString(";,");
            if (next == null) {
                throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
            }
            final ArrayList headerValues = new ArrayList();
            final StringBuffer headerValue = new StringBuffer(next);
            headerValues.add(next);
            boolean directive = false;
            char c = tokenizer.getChar();
            while (c == ';') {
                next = tokenizer.getString(";,=:");
                if (next == null) {
                    throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                }
                c = tokenizer.getChar();
                while (c == ':') {
                    c = tokenizer.getChar();
                    if (c != '=') {
                        final String restOfNext = tokenizer.getToken(";,=:");
                        if (restOfNext == null) {
                            throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                        }
                        next = String.valueOf(next) + ":" + c + restOfNext;
                        c = tokenizer.getChar();
                    }
                    else {
                        directive = true;
                    }
                }
                if (c != ';' && c != ',' && c != '\0') {
                    continue;
                }
                headerValues.add(next);
                headerValue.append(";").append(next);
            }
            final ManifestElement manifestElement = new ManifestElement();
            manifestElement.value = headerValue.toString();
            manifestElement.valueComponents = headerValues.toArray(new String[headerValues.size()]);
            while (c == '=' || c == ':') {
                while (c == ':') {
                    c = tokenizer.getChar();
                    if (c != '=') {
                        final String restOfNext2 = tokenizer.getToken("=:");
                        if (restOfNext2 == null) {
                            throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                        }
                        next = String.valueOf(next) + ":" + c + restOfNext2;
                        c = tokenizer.getChar();
                    }
                    else {
                        directive = true;
                    }
                }
                final String val = tokenizer.getString(";,");
                if (val == null) {
                    throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                }
                try {
                    if (directive) {
                        manifestElement.addDirective(next, val);
                    }
                    else {
                        manifestElement.addAttribute(next, val);
                    }
                    directive = false;
                }
                catch (final Exception ex) {
                    throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                }
                c = tokenizer.getChar();
                if (c != ';') {
                    continue;
                }
                next = tokenizer.getToken("=:");
                if (next == null) {
                    throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
                }
                c = tokenizer.getChar();
            }
            headerElements.add(manifestElement);
            if (c == ',') {
                continue;
            }
            if (c != '\0') {
                throw new BundleException(NLS.bind("Invalid manifest header {0}: \"{1}\"", header, value), 3);
            }
            final int size = headerElements.size();
            if (size == 0) {
                return null;
            }
            final ManifestElement[] result = headerElements.toArray(new ManifestElement[size]);
            return result;
        }
    }
    
    public static String[] getArrayFromList(final String stringList) {
        final String[] result = getArrayFromList(stringList, ",");
        return (String[])((result.length == 0) ? null : result);
    }
    
    public static String[] getArrayFromList(final String stringList, final String separator) {
        if (stringList == null || stringList.trim().length() == 0) {
            return new String[0];
        }
        final ArrayList list = new ArrayList();
        final StringTokenizer tokens = new StringTokenizer(stringList, separator);
        while (tokens.hasMoreTokens()) {
            final String token = tokens.nextToken().trim();
            if (token.length() != 0) {
                list.add(token);
            }
        }
        return list.toArray(new String[list.size()]);
    }
    
    public static Map parseBundleManifest(final InputStream manifest, Map headers) throws IOException, BundleException {
        if (headers == null) {
            headers = new HashMap();
        }
        BufferedReader br;
        try {
            br = new BufferedReader(new InputStreamReader(manifest, "UTF8"));
        }
        catch (final UnsupportedEncodingException ex) {
            br = new BufferedReader(new InputStreamReader(manifest));
        }
        try {
            String header = null;
            final StringBuffer value = new StringBuffer(256);
            boolean firstLine = true;
            while (true) {
                final String line = br.readLine();
                if (line == null || line.length() == 0) {
                    if (!firstLine) {
                        headers.put(header, value.toString().trim());
                        break;
                    }
                    break;
                }
                else if (line.charAt(0) == ' ') {
                    if (firstLine) {
                        throw new BundleException(NLS.bind("The manifest line \"{0}\" has an invalid leading space '' '' character.", line), 3);
                    }
                    value.append(line.substring(1));
                }
                else {
                    if (!firstLine) {
                        headers.put(header, value.toString().trim());
                        value.setLength(0);
                    }
                    final int colon = line.indexOf(58);
                    if (colon == -1) {
                        throw new BundleException(NLS.bind("The manifest line \"{0}\" is invalid; it has no colon '':'' character after the header key.", line), 3);
                    }
                    header = line.substring(0, colon).trim();
                    value.append(line.substring(colon + 1));
                    firstLine = false;
                }
            }
        }
        finally {
            try {
                br.close();
            }
            catch (final IOException ex2) {}
        }
        try {
            br.close();
        }
        catch (final IOException ex3) {}
        return headers;
    }
    
    @Override
    public String toString() {
        final Enumeration attrKeys = this.getKeys();
        final Enumeration directiveKeys = this.getDirectiveKeys();
        if (attrKeys == null && directiveKeys == null) {
            return this.value;
        }
        final StringBuffer result = new StringBuffer(this.value);
        if (attrKeys != null) {
            while (attrKeys.hasMoreElements()) {
                final String key = attrKeys.nextElement();
                this.addValues(false, key, this.getAttributes(key), result);
            }
        }
        if (directiveKeys != null) {
            while (directiveKeys.hasMoreElements()) {
                final String key = directiveKeys.nextElement();
                this.addValues(true, key, this.getDirectives(key), result);
            }
        }
        return result.toString();
    }
    
    private void addValues(final boolean directive, final String key, final String[] values, final StringBuffer result) {
        if (values == null) {
            return;
        }
        for (int i = 0; i < values.length; ++i) {
            result.append(';').append(key);
            if (directive) {
                result.append(':');
            }
            result.append("=\"").append(values[i]).append('\"');
        }
    }
}
